/*
**++
**  FACILITY:	NETLIB test client
**
**  ABSTRACT:	Echo client.
**
**  MODULE DESCRIPTION:
**
**  	This is a test client for NETLIB.  It connects to an "echo" server
**  	via TCP, which is supposed to echo back any characters sent to it.
**
**  	We do character-at-a-time I/O, which isn't very efficient, but
**  	it's a good test of NETLIB's asynchronous I/O capabilities.
**
**  	N.B.: This program expects that SYS$INPUT and SYS$OUTPUT point
**  	      to terminal devices.
**
**  AUTHOR: 	    M. Madison
**  	    	    COPYRIGHT  1994, MADGOAT SOFTWARE.  ALL RIGHTS RESERVED.
**
**  CREATION DATE:  17-NOV-1994
**
**  MODIFICATION HISTORY:
**
**  	17-NOV-1994 V1.0    Madison 	Initial coding.
**--
*/
#include "netlib_dir:netlibdef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <descrip.h>
#include <stsdef.h>
#include <ssdef.h>
#include <iodef.h>
#include <ttdef.h>
#include <tt2def.h>
#include <lib$routines.h>
#include <starlet.h>
#ifdef __DECC
#include <builtins.h>
#else
#pragma builtins
#endif

struct dev_char {               /* Device characteristics block       */
  unsigned char class;
  unsigned char type;
  short int buffer_size;
  unsigned int basic_chars;
  unsigned int extended_chars;
};

void hd(char *info, unsigned char *ptr, int count)
{
    char s[80], *p;
    int line, col;

    for (line = 0; line < count; line += 8)
    {
        p = s;

        p += sprintf(p, "%s", info);

        p += sprintf(p, "%04x  ", line);

        for (col = 0; col < 8; col++)
        {
            if (line + col < count)
                p += sprintf(p, "%02x ", ptr[line + col]);
            else
                p += sprintf(p, "   ");
        }

        *p++ = ' ';

        for (col = 0; col < 8; col++)
        {
            if (line + col >= count)
                break;

            if (ptr[line + col] >= ' ' && ptr[line + col] < 127)
                *p++ = ptr[line + col];
            else
                *p++ = '.';
        }

        *p = '\0';

        printf("%s\n", s);
    }
}

#define INIT_SDESC(dsc, len, ptr) {(dsc).dsc$b_dtype = DSC$K_DTYPE_T;\
    (dsc).dsc$b_class = DSC$K_CLASS_S; (dsc).dsc$w_length = (len);\
    (dsc).dsc$a_pointer = (ptr);}

/*
**  This is the default port we use (matches the accompanying server program).
*/
    const unsigned short ECHO_PORT = 2000;

/*
**  Other statics we need because we have to pass these values by
**  reference
*/
    static unsigned int sinsize = sizeof(struct SINDEF);
    static unsigned int adrsize = sizeof(struct INADDRDEF);
    static unsigned int which_dns = NETLIB_K_LOOKUP_DNS;
    static unsigned int which_ht  = NETLIB_K_LOOKUP_HOST_TABLE;
/*
**  I/O request
*/
#define IORBUFSize 16384
    typedef struct IORequest {
    	struct IORequest *flink, *blink;
    	struct NETLIBIOSBDEF iosb;
    	struct dsc$descriptor bufdsc;
    	char buf[IORBUFSize];
    } IORequest;
/*
**  Queue
*/
    typedef struct Queue {void *head, *tail;} Queue;
/*
**  Forward declarations
*/
    unsigned int main(int argc, char *argv[]);
    void Terminal_Read_AST(IORequest *ior);
    void Network_Read_AST(IORequest *ior);
    void Write_AST(IORequest *ior);
    IORequest *Allocate_IOR(void);
    IORequest *Get_IOR(void);
/*
**  OWN storage
*/
    static Queue Free_Queue = {&Free_Queue, &Free_Queue};
    static $DESCRIPTOR(input_device, "SYS$INPUT");
    static $DESCRIPTOR(output_device, "SYS$OUTPUT");
    static $DESCRIPTOR(help, "Usage: ECHOCLIENT hostname [port-number]");
    static void *network_context;
    static unsigned short network_ef;
    static unsigned long network_ef_mask;
    static unsigned char network_buffer[IORBUFSize];
    static unsigned char network_buffer_n = 0;
    static unsigned short input_channel;
    static unsigned short input_ef;
    static unsigned long input_ef_mask;
    static unsigned char input_buffer[IORBUFSize];
    static unsigned char input_buffer_n = 0;
    static unsigned short output_channel;
    unsigned short done_ef;
    static unsigned long done_ef_mask;
    static unsigned long ef_bits;
    static unsigned int completion_status = SS$_NORMAL;
    const unsigned int terminator_mask[2] = {0, 0};
/*
 * We're gonna change some of the characteristics of the terminal so I'll
 * declare a buffer for the terminal's characteristics before the changes
 * and another for after the changes. When the program exits the terminal's
 * characteristics will be reset the way they were before they were changed.
 */

static struct dev_char otinput, ntinput;
static struct dev_char otoutput, ntoutput;


#include "Pilot.h"

#include "ssh.h"
#include "ui.h"
#include "rand.h"
#include "erf.h"

/* We know, it's not safe, but we know how to use it correctly. */
#define min(a,b) ((a) < (b) ? (a) : (b))

void *xmalloc(size_t size)
{
    void *buf;
    
    if ((buf = malloc(size)) == NULL)
	ssh_fatal("malloc");
    
    return buf;
}

char *xstrdup(const char *s)
{
    char *r;
    
    r = xmalloc(strlen(s) + 1);
    strcpy(r, s);

    return r;
}

static void ssh_initprefs(ssh_prefs *p)
{
    char *cp;

    p->host = NULL;
    p->port = 22;
    p->username = xstrdup(getenv("USER"));
    for (cp = p->username; *cp; cp++)
	*cp = tolower(*cp);
    p->ciphers = ((1 << SSH_CIPHER_IDEA) | (1 << SSH_CIPHER_3DES));
    p->ttymodes = NULL;
    p->ttymodelen = 0;
}

static void ssh_clearprefs(ssh_prefs *p)
{
    if (p->host) MemPtrFree(p->host);
    p->port = 0;
    if (p->username) MemPtrFree(p->username);
    if (p->ttymodes) MemPtrFree(p->ttymodes);
    ssh_initprefs(p);
}

void connected(Erf erf, void *erfp)
{
    IORequest *ior;
    unsigned int status;

    ssh_info("Entering interactive mode");

    fprintf(stderr, "\n\n");

    ntinput = otinput;
    ntinput.basic_chars |= TT$M_WRAP | /*TT$M_PASSALL |*/ TT$M_NOECHO | TT$M_HOSTSYNC;
    ntinput.basic_chars &= ~(/*TT$M_HOSTSYNC |*/ TT$M_TTSYNC /*| TT$M_READSYNC*/);
    ntinput.extended_chars |= TT2$M_PASTHRU;
    ntinput.extended_chars &= ~(TT2$M_LOCALECHO);
    status = sys$qiow (0, input_channel, IO$_SETMODE, 0, 0, 0,
		       &ntinput, sizeof(ntinput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) exit(status);

    ntoutput = otoutput;
    ntoutput.basic_chars |= TT$M_WRAP | /*TT$M_PASSALL |*/ TT$M_NOECHO | TT$M_HOSTSYNC;
    ntoutput.basic_chars &= ~(/*TT$M_HOSTSYNC |*/ TT$M_TTSYNC /*| TT$M_READSYNC*/);
    ntoutput.extended_chars |= TT2$M_PASTHRU;
    ntoutput.extended_chars &= ~(TT2$M_LOCALECHO);
    status = sys$qiow (0, output_channel, IO$_SETMODE, 0, 0, 0,
		       &ntoutput, sizeof(ntoutput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) exit(status);

/*
**  Start a terminal read
*/
    ior = Get_IOR();
    if (ior == 0) exit(SS$_INSFMEM);
    status = sys$qio(0, input_channel,
    	    	IO$_READVBLK|IO$M_NOECHO|IO$M_NOFILTR,
    	    	&ior->iosb, Terminal_Read_AST, ior,
    	    	ior->buf, 1, 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) exit(status);
}

static ssh_prefs gsshprefs;
static void *sshstate = NULL;

int sockfd;

static int gerrp;
Erf erf, erfp;

long sock_write(unsigned char *data, long len, void *state_,
    Erf erf, void *erfp)
{
    IORequest *ior;
    long totret = 0;
    unsigned int status;

    if (len < 0) {
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);

	totret = -1;
	goto sockwrite_out;
    }

    totret = len;

    while (len > 0) {
	int this_len = min(len, IORBUFSize);

/*
**  Write the data to the server.
*/
	ior = Get_IOR();
	if (ior == 0) status = SS$_INSFMEM;
	else
	{
	    ior->bufdsc.dsc$w_length = this_len;
	    memcpy(ior->buf, data, this_len);
	    status = netlib_write(&network_context, &ior->bufdsc,
				  0, 0, &ior->iosb, Write_AST, ior);
	}

	len -= this_len;

	if (!$VMS_STATUS_SUCCESS(status)) {
	    unsigned long ef_bits_to_check;
	    sys$readef(input_ef, ef_bits_to_check);
	    if (!(ef_bits_to_check | done_ef)) completion_status = status;
#ifdef DEBUG
	    lib$signal(SS$_DEBUG);
#endif
	    sys$setef(done_ef);
	    totret = -1;
	    goto sockwrite_out;
	}
    }

sockwrite_out:
    return totret;
}

/* This function is called when data arrives */
void ui_dispatch(unsigned char *data, long len, void *state_, Erf erf,
    void *erfp)
{
    IORequest *ior;
    unsigned int status;

    if (len < 0) {
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);

	/* The connection was closed; pass on the info */
	goto ui_dispatch_out;
    }

    while (len > 0) {
	int this_len = min(len, IORBUFSize);

	ior = Get_IOR();
	if (ior == 0) status = SS$_INSFMEM;
	else 
	{
	    memcpy(ior->buf, data, this_len);

	    status = sys$qio(0, output_channel, IO$_WRITEVBLK,
			     &ior->iosb, Write_AST, ior, ior->buf, this_len,
			     0, 0, 0, 0);
	}

	len -= this_len;

	if (!$VMS_STATUS_SUCCESS(status)) {
	    unsigned long ef_bits_to_check;
	    sys$readef(input_ef, ef_bits_to_check);
	    if (!(ef_bits_to_check | done_ef)) completion_status = status;
#ifdef DEBUG
	    lib$signal(SS$_DEBUG);
#endif
	    sys$setef(done_ef);
	}
    }
	    
ui_dispatch_out:
    ;
}

/*
 *      e x i t _ h a n d l e r
 */
static void exit_handler ()
{
    sys$setast (0);
    
    sys$dassgn(input_channel);
    sys$dassgn(output_channel);
    sys$assign(&input_device, &input_channel, 0, 0);
    sys$assign(&output_device, &output_channel, 0, 0);
    sys$qiow (0, input_channel, IO$_SETMODE, 0, 0, 0,
	      &otinput, sizeof(otinput), 0, 0, 0, 0);
    sys$qiow (0, output_channel, IO$_SETMODE, 0, 0, 0,
	      &otoutput, sizeof(otoutput), 0, 0, 0, 0);
    sys$dassgn(input_channel);
    sys$dassgn(output_channel);
}

void sighand(int signo)
{
    signal(SIGINT, SIG_DFL);

    fprintf(stderr, "Exiting on interrupt\n");

    exit(0);
}

unsigned int main(int argc, char *argv[])
{
    char randfilename[32];
    IORequest *ior;
    struct SINDEF remsin;
    struct dsc$descriptor dsc;
    unsigned int status, remlen;
    unsigned short port, namelen;
    char name[1024], buf[1024];
    int i;
    static int excode;
    struct exit_handler_block 
    {
        struct exit_handler_block *flink;
        void (*exit_routine)();
        int arg_count;
        int *status_address;
        int exit_status;
    } exhblk = {0, exit_handler, 1, &excode, 0};

    if (argc < 2)
    {
	fprintf(stderr, "Usage: %s host [port] [user]\n", argv[0]);
	exit(1);
    }

    ssh_initprefs(&gsshprefs);

    gsshprefs.host = xstrdup(argv[1]);
    if (argc >= 3)
	gsshprefs.port = atoi(argv[2]);
    if (argc >= 4)
	gsshprefs.username = xstrdup(argv[3]);

    /* Load up with some randomness; use the saved file and the memopad */

    RAND_file_name(randfilename, 32);
    RAND_load_file(randfilename, 0x7fffffffL);

/*
**  Get channels to the input and output devices
*/
    status = sys$assign(&input_device, &input_channel, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    status = sys$assign(&output_device, &output_channel, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

/*
**  Save terminal characteristics and set up an exit function
**  to restore them on exit.
*/    
    status = sys$qiow (0, input_channel, IO$_SENSEMODE, 0, 0, 0,
		       &otinput, sizeof(otinput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);
    status = sys$qiow (0, output_channel, IO$_SENSEMODE, 0, 0, 0,
		       &otoutput, sizeof(otoutput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);

    status = sys$dclexh (&exhblk);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);

    signal(SIGINT, sighand);

    ntinput = otinput;
    ntinput.basic_chars |= TT$M_NOECHO;
#if 0
    ntinput.extended_chars &= ~(TT2$M_LOCALECHO);
#endif
    status = sys$qiow (0, input_channel, IO$_SETMODE, 0, 0, 0,
		       &ntinput, sizeof(ntinput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);

    ntoutput = otoutput;
    ntoutput.basic_chars |= TT$M_NOECHO;
#if 0
    ntoutput.extended_chars &= ~(TT2$M_LOCALECHO);
#endif
    status = sys$qiow (0, output_channel, IO$_SETMODE, 0, 0, 0,
		       &ntoutput, sizeof(ntoutput), 0, 0, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);

/*
**  Allocate a network socket and connect to the server
*/
    status = netlib_socket(&network_context);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    INIT_SDESC(dsc, strlen(gsshprefs.host), gsshprefs.host);
    port = gsshprefs.port;
    status = netlib_connect_by_name(&network_context, &dsc, &port);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

/*
**  Allocate event flags for signaling to the main loop.
*/
    status = lib$get_ef(&input_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    input_ef_mask = 1<<(input_ef & 31);
    status = sys$clref(input_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    status = lib$get_ef(&network_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    network_ef_mask = 1<<(network_ef & 31);
    status = sys$clref(network_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    status = lib$get_ef(&done_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    done_ef_mask = 1<<(done_ef & 31);
    status = sys$clref(done_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    ef_bits = input_ef_mask | network_ef_mask | done_ef_mask;

/*
**  Let the user know who we're connected to
*/
    status = netlib_getpeername(&network_context, &remsin, &sinsize, &remlen);
    if ($VMS_STATUS_SUCCESS(status)) {
    	INIT_SDESC(dsc, sizeof(name), name);
    	status = netlib_address_to_name(&network_context, &which_dns,
    	    	    &remsin.sin_x_addr, &adrsize, &dsc, &namelen);
    	if (!$VMS_STATUS_SUCCESS(status))
    	    status = netlib_address_to_name(&network_context, &which_ht,
    	    	    &remsin.sin_x_addr, &adrsize, &dsc, &namelen);
    	if (!$VMS_STATUS_SUCCESS(status))
    	    status = netlib_addrtostr(&remsin.sin_x_addr, &dsc, &namelen);
    	if ($VMS_STATUS_SUCCESS(status)) {
    	    i = sprintf(buf, "Connected to host %.*s, port %d.",
    	    	namelen, name, netlib_ntoh_word(&remsin.sin_w_port));
    	    INIT_SDESC(dsc, i, buf);
    	    lib$put_output(&dsc);
    	    dsc.dsc$w_length = 0;
    	    lib$put_output(&dsc);
    	}
    }

/*
**  Pre-load the free queue
*/
    for (i = 0; i < 16; i++) {
    	ior = Allocate_IOR();
    	if (ior == 0) break;
    	_INSQUE(ior, Free_Queue.tail);
    }

    /* Set up an SSH connection */
    sshstate = ssh_init(sock_write, NULL, erf, erfp);
    if (!sshstate) exit(1);

    /* Set the preferences */
    ssh_set_prefs(&gsshprefs, sshstate, erf, erfp);

    {
	char *termname = getenv("TERM");

	if (termname)
	{
	    char *t2 = xstrdup(termname);
	    char *p;

	    if ((p = strchr(t2, '-')) != NULL)
		*p = '\0';

	    ssh_setterm(t2, 0, 0, 0, 0, sshstate, erf, erfp);
	    
	    free(t2);
	}
    }

    /* Set the dispatch path */
    ssh_set_dispatch(ui_dispatch, NULL, sshstate, erf, erfp);

/*
**  Start a network read
*/
    ior = Get_IOR();
    if (ior == 0) return SS$_INSFMEM;
    status = netlib_read(&network_context, &ior->bufdsc,
    	    	    	  0, 0, 0, 0, &ior->iosb,
    	    	    	  Network_Read_AST, ior);
    if (!$VMS_STATUS_SUCCESS(status)) {
    	sys$cancel(input_channel);
    	return SS$_INSFMEM;
    }

/*
**  Let the ASTs do most of the work.  They will set appropriate event
**  flags when it's time for this loop to do some external work, or when
**  it's time to shut down.
*/
    while (1) {
	unsigned long ef_bits_to_check;
	status = sys$wflor(input_ef,ef_bits);
	status = sys$readef(input_ef,&ef_bits_to_check);
	if (ef_bits_to_check & input_ef_mask) {
	    sys$setast(0);
/*
**  Send data downlink through SSH layer
*/
#if 1
	    ssh_write(input_buffer, input_buffer_n, sshstate, erf, &gerrp);
#else
	    sock_write(input_buffer, input_buffer_n, sshstate, erf, &gerrp);
#endif
	    input_buffer_n = 0;
	    sys$clref(input_ef);
	    sys$setast(1);
	}
	if (ef_bits_to_check & done_ef_mask) {
	    break;
	}
    }


/*
**  Close down the network socket and terminal channels.
*/
    netlib_shutdown(&network_context);
    netlib_close(&network_context);
    sys$dassgn(input_channel);
    sys$dassgn(output_channel);

    /* Save our randomness state */
    RAND_file_name(randfilename, 32);
    RAND_write_file(randfilename);

    ssh_clearprefs(&gsshprefs);

    return completion_status;
}



/*
**++
**  ROUTINE:	Terminal_Read_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Read completion AST for terminal input.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Terminal_Read_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Terminal_Read_AST (IORequest *ior) {

    unsigned int status;
    static enum { NOTHING, CR_SEEN, TILDE_SEEN } keystate = NOTHING;
    register unsigned int c = -1;

    if (ior->iosb.iosb_w_count != 0) {
/*
**  Exit on error
*/
	if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)) {
	    unsigned long ef_bits_to_check;
	    sys$readef(input_ef, ef_bits_to_check);
	    if (!(ef_bits_to_check | done_ef))
		completion_status = ior->iosb.iosb_w_status;
#ifdef DEBUG
	    lib$signal(SS$_DEBUG);
#endif
	    sys$setef(done_ef);
	    return;
	}

	c = ior->buf[0];
    }

/*
**  Start a new read
*/
    status = sys$qio(0, input_channel,
		     IO$_READVBLK|IO$M_NOECHO|IO$M_NOFILTR,
		     &ior->iosb, Terminal_Read_AST, ior,
		     ior->buf, 1, 0, terminator_mask, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, ef_bits_to_check);
	if (!(ef_bits_to_check | done_ef))
	    completion_status = ior->iosb.iosb_w_status;
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);
    	return;
    }

    if (c >= 0) {
	/*fprintf(stderr, "got 0x%x\n", c);*/

	switch (keystate)
	{
	case NOTHING:
	    if (c == '\r' || c == '\n')
		keystate = CR_SEEN;
	    break;
	case CR_SEEN:
	    if (c == '~')
		keystate = TILDE_SEEN;
	    else
		keystate = NOTHING;
	    break;
	case TILDE_SEEN:
	    if (c == '.')
	    {
		fprintf(stderr, "Connection closed\n");
#ifdef DEBUG
		lib$signal(SS$_DEBUG);
#endif
		sys$setef(done_ef);
	    }
	    else
		keystate = NOTHING;
	    break;
	}
	input_buffer[input_buffer_n++] = c;
	sys$setef(input_ef);
    }
} /* Terminal_Read_AST */

/*
**++
**  ROUTINE:	Network_Read_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Completion AST for network read.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Network_Read_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Network_Read_AST (IORequest *ior) {

    unsigned int status;
    int len;

/*
**  Exit on error
*/
    if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, ef_bits_to_check);
    	if (!(ef_bits_to_check | done_ef))
	    completion_status = ior->iosb.iosb_w_status;
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);
    	return;
    }

/* 
**  Send data up through ssh layer 
*/  
#if 1
    ssh_dispatch(ior->buf, 1, sshstate, erf, erfp);
#else
    ui_dispatch(ior->buf, 1, sshstate, erf, erfp);
#endif
    
/*
**  Start a new read from the network
*/
    status = netlib_read(&network_context, &ior->bufdsc,
			 0, 0, 0, 0, &ior->iosb,
			 Network_Read_AST, ior);
    if (!$VMS_STATUS_SUCCESS(status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, ef_bits_to_check);
    	if (!(ef_bits_to_check | done_ef))
	    completion_status = ior->iosb.iosb_w_status;
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);
    	return;
    }
} /* Network_Read_AST */

unsigned long ior_allocation_count = 0;	/* Statistics */
unsigned long ior_use_count = 0;	/* Statistics */

/*
**++
**  ROUTINE:	Write_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Completion AST for terminal and network writes.  It checks
**  for successful completion and frees the I/O request block.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Write_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Write_AST (IORequest *ior) {

    if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, ef_bits_to_check);
    	if (!(ef_bits_to_check | done_ef))
	    completion_status = ior->iosb.iosb_w_status;
#ifdef DEBUG
	lib$signal(SS$_DEBUG);
#endif
	sys$setef(done_ef);
    	return;
    }

    _INSQUE(ior, Free_Queue.tail);
    ior_use_count--;


} /* Write_AST */

/*
**++
**  ROUTINE:	Allocate_IOR
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Allocates an IORequest.
**
**  RETURNS:    IORequest *
**
**  PROTOTYPE:
**
**  	Allocate_IOR()
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	0 = allocation failure
**
**  SIDE EFFECTS:   	None.
**
**--
*/
IORequest *Allocate_IOR (void) {

    IORequest *ior;
    unsigned int status;
    const unsigned int iorsize = sizeof(IORequest);

    status = lib$get_vm(&iorsize, &ior);
    if (!$VMS_STATUS_SUCCESS(status)) return 0;

    memset(ior, 0, iorsize);
    INIT_SDESC(ior->bufdsc, 1, ior->buf);

    ior_allocation_count++;

    return ior;

} /* Allocate_IOR */

/*
**++
**  ROUTINE:	Get_IOR
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Gets a free IORequest, either off the free queue or by
**  allocating one.
**
**  RETURNS:	IORequest *
**
**  PROTOTYPE:
**
**  	Get_IOR()
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	0 = couldn't get one
**
**  SIDE EFFECTS:   	None.
**
**--
*/
IORequest *Get_IOR (void) {

    IORequest *ior;

    /* Let's watch out, it looks like things get a little overwritten if
       we don't do the extra check on Free_Queue.head... */
    if (Free_Queue.head == &Free_Queue
	/* Only return code 2 means there was nothing in the queue... */
	|| _REMQUE(Free_Queue.head, (void **) &ior) == 2) 
	ior = Allocate_IOR();

    if (ior)
	ior->bufdsc.dsc$w_length = 1;

    ior_use_count++;

    return ior;
} /* Get_IOR */
